@@ -1,23 +1,21 @@
-/* Copyright 2014-2023 Gentoo Foundation
- * Distributed under the terms of the GNU General Public License v2
- *
+/*
  * Wrapper for coreutils install to preserve extended attributes.
  *
- * Copyright 2014 Anthony G. Basile - <blueness@gentoo.org>
- * Copyright 2014 Mike Frysinger    - <vapier@gentoo.org>
- *
- * This program is free software: you can redistribute it and/or modify
- * it under the terms of the GNU General Public License as published by
- * the Free Software Foundation, either version 3 of the License, or
- * (at your option) any later version.
+ * Copyright 2014-2026 Gentoo Authors
  *
+ * This program is free software; you can redistribute it and/or
+ * modify it under the terms of the GNU General Public License
+ * as published by the Free Software Foundation; either version 2
+ * of the License, or (at your option) any later version.
+
  * This program is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  * GNU General Public License for more details.
- *
+
  * You should have received a copy of the GNU General Public License
- * along with this program.  If not, see <http://www.gnu.org/licenses/>.
+ * along with this program; if not, see
+ * <https://www.gnu.org/licenses/>.
  */
 
 #include <ctype.h>
@@ -30,6 +28,7 @@
 #include <stdlib.h>
 #include <signal.h>
 #include <spawn.h>
+#include <stdbool.h>
 #include <string.h>
 #include <sys/types.h>
 #include <sys/wait.h>
@@ -41,14 +40,7 @@
 #define REAL_INSTALL "/usr/bin/install"
 #endif
 
-static char *
-xstrdup(const char *s)
-{
-	char *ret = strdup(s);
-	if (ret == NULL)
-		err(1, "strdup() failed");
-	return ret;
-}
+#define DEFAULT_EXCLUDE "btrfs.*\0security.*\0trusted.*\0system.nfs4_acl"
 
 static void *
 xmalloc(size_t size)
@@ -110,13 +102,12 @@ xsetxattr(const char *path, char *list, char *value, size_t size)
 	return ret;
 }
 
-char *exclude;      /* strings of excluded xattr names              */
-size_t len_exclude; /* length of the string of excluded xattr names */
+static const char *exclude; /* strings of excluded xattr names */
+static size_t len_exclude;  /* length of the string of excluded xattr names */
 
 static void
 copyxattr(const char *source, const char *target)
 {
-	ssize_t i, j;           /* counters to walk through strings                   */
 	ssize_t lsize, xsize;   /* size in bytes of the list of xattrs and the values */
 	char *lxattr;                  /* string of xattr names                       */
 	static char *value = NULL ;    /* string of an xattr name's value             */
@@ -138,40 +129,33 @@ copyxattr(const char *source, const char *target)
 	lxattr = xmalloc(lsize);
 	lsize = xlistxattr(source, lxattr, lsize);
 
-	i = 0;
-	while (1) {
-		while (lxattr[i++] == 0)
-			continue;
+	for (size_t i = 0;;) {
+		while (i < lsize && lxattr[i] == '\0')
+			++i;
 		if (i >= lsize)
 			break;
 
-		j = 0;
-		while (1) {
-			while (exclude[j++] == 0)
-				continue;
+		for (size_t j = 0;;) {
+			while (j < len_exclude && exclude[j] == '\0')
+				++j;
 			if (j >= len_exclude)
 				break;
-			if (!fnmatch(&exclude[j - 1], &lxattr[i - 1], 0))
+			if (!fnmatch(exclude + j, lxattr + i, 0))
 				goto skip;
-			while (exclude[j++] != 0)
-				continue;
-			if (j >= len_exclude)
-				break;
+			while (j < len_exclude && exclude[j] != '\0')
+				++j;
 		}
 
-		xsize = xgetxattr(source, &lxattr[i-1], 0, 0);
+		xsize = xgetxattr(source, lxattr + i, 0, 0);
 		if (xsize > value_size) {
 			value_size = xsize;
 			value = xrealloc(value, value_size);
 		}
-		xgetxattr(source, &lxattr[i-1], value, xsize);
-		xsetxattr(target, &lxattr[i-1], value, xsize);
-
+		xgetxattr(source, lxattr + i, value, xsize);
+		xsetxattr(target, lxattr + i, value, xsize);
  skip:
-		while (lxattr[i++] != 0)
-			continue;
-		if (i >= lsize)
-			break;
+		while (i < lsize && lxattr[i] != '\0')
+			++i;
 	}
 
 	/* No need to free(value) on return because its static and we */
@@ -198,40 +182,32 @@ spawn(char *path, char *argv[])
 int
 main(int argc, char* argv[])
 {
-	int i;
-	int status;                    /* exit status of child "install" process                       */
+	/* If this is set it means Portage has switched the working directory
+	 * and expects us to switch back. */
+	char *portage_helper_cwd = getenv("__PORTAGE_HELPER_CWD");
+	if (portage_helper_cwd && chdir(portage_helper_cwd))
+		err(1, "failed to chdir %s", portage_helper_cwd);
 
-	char** argv_copy;              /* copy argv to avoid mangling by getopt_long()                 */
+	argv[0] = getenv("REAL_INSTALL") ?: REAL_INSTALL;
 
-	int opts_directory = 0;        /* if -d was given, then all arguments are directories          */
-	int opts_target_directory = 0; /* if -t was given, then we're installing to a target directory */
-	int target_is_directory = 0;   /* is the target a directory?                                   */
+	if (spawn(argv[0], argv) < 0)
+		err(1, "failed to spawn install");
 
-	int first, last;               /* argv indices of the first file/directory and last            */
-	char *target = NULL;           /* the target file or directory                                 */
-	char *path;                    /* path to the target file                                      */
-
-	struct stat s;                 /* test if a file is a regular file or a directory              */
-
-	char *portage_xattr_exclude;   /* strings of excluded xattr names from $PORTAGE_XATTR_EXCLUDE  */
-
-	portage_xattr_exclude = getenv("PORTAGE_XATTR_EXCLUDE");
-	if (portage_xattr_exclude == NULL)
-		exclude = xstrdup("btrfs.* security.* trusted.* system.nfs4_acl");
-	else
-		exclude = xstrdup(portage_xattr_exclude);
-
-	len_exclude = strlen(exclude);
-
-	/* We convert exclude[] to an array of concatenated NUL terminated
-	 * strings.  Also, no need to free(exclude) before we exit().
-	 */
-	char *p = exclude;
-	char *pend = p + len_exclude;
-	while (p != pend) {
-		if (isspace(*p))
-			*p = '\0';
-		p++;
+	char *portage_xattr_exclude = getenv("PORTAGE_XATTR_EXCLUDE");
+	if (portage_xattr_exclude) {
+		exclude = portage_xattr_exclude;
+		len_exclude = strlen(exclude);
+		/* We convert exclude[] to an array of concatenated NUL terminated
+		 * strings.  Also, no need to free(exclude) before we exit().
+		 */
+		char *pend = portage_xattr_exclude + len_exclude;
+		for (char *p = portage_xattr_exclude; p != pend; ++p)
+			if (isspace(*p))
+				*p = '\0';
+	}
+	else {
+		exclude = DEFAULT_EXCLUDE;
+		len_exclude = sizeof(DEFAULT_EXCLUDE) - 1;
 	}
 
 	opterr = 0; /* we skip many legitimate flags, so silence any warning */
@@ -249,19 +225,13 @@ main(int argc, char* argv[])
 		{                     0,                 0, 0,  0 }
 	};
 
-	/* We copy argv[] because getopts_long() mangles the order of the arguments.
-	 * We pass the original argv[] to install in the fork() while we use
-	 * argv_copy[] for the copying of the xattrs since optind assumes a reorder
-	 * parameter list.
-	 */
-	argv_copy = (char **)malloc(argc*sizeof(char *));
-
-	for (i = 0; i < argc; i++)
-		argv_copy[i] = strdup(argv[i]);
+	bool opts_directory = false;
+	bool opts_target_directory = false;
+	bool target_is_directory = false;
+	char *target = NULL;
 
 	while (1) {
-		int option_index;
-		int c = getopt_long(argc, argv_copy, "dt:g:m:o:S:Zb", long_options, &option_index);
+		int c = getopt_long(argc, argv, "dt:g:m:o:S:Zb", long_options, NULL);
 
 		if (c == -1)
 			break;
@@ -279,11 +249,11 @@ main(int argc, char* argv[])
 				break;
 
 			case 'd':
-				opts_directory = 1;
+				opts_directory = true;
 				break;
 
 			case 't':
-				opts_target_directory = 1;
+				opts_target_directory = true;
 				target = optarg;
 				break;
 
@@ -292,27 +262,7 @@ main(int argc, char* argv[])
 		}
 	}
 
-	first = optind;
-	last = argc - 1;
-
-	/* Do we need to chdir to OLDPWD?  This is required when we are called my a
-	 * wrapper like ${__PORTAGE_HELPER_PATH} which then passes its directory as
-	 * $PWD and the source directory from which it was called as $OLDPWD.  But
-	 * we want the system install to run in the source directory, ie $OLDPWD,
-	 * so we chdir to it.  Currently we assume that if __PORTAGE_HELPER_PATH
-	 * is set, then we chdir to oldpwd.
-	 */
-	char *oldpwd = getenv("OLDPWD");
-	char *portage_helper_path = getenv("__PORTAGE_HELPER_PATH");
-	if (portage_helper_path)
-		if (!oldpwd || chdir(oldpwd) != 0)
-			err(1, "failed to chdir %s", oldpwd);
-
-	argv[0] = getenv("REAL_INSTALL") ?: REAL_INSTALL;
-
-	if (spawn(argv[0], argv) < 0)
-		err(1, "failed to spawn install");
-
+	int status;
 	wait(&status);
 
 	/* Are there enough files/directories on the cmd line to
@@ -320,6 +270,8 @@ main(int argc, char* argv[])
 	 * arguments or with just --help.  In which case there is
 	 * nothing for the parent to do.
 	 */
+	int first = optind;
+	int last = argc - 1;
 	if (first >= last)
 		goto done;
 
@@ -328,7 +280,8 @@ main(int argc, char* argv[])
 		goto done;
 
 	if (!opts_target_directory) {
-		target = argv_copy[last];
+		struct stat s;
+		target = argv[last];
 		if (stat(target, &s) != 0) {
 			err(1, "failed to stat %s", target);
 			return EXIT_FAILURE;
@@ -336,7 +289,7 @@ main(int argc, char* argv[])
 		target_is_directory = S_ISDIR(s.st_mode);
 	} else {
 		/* target was set above with the -t option */
-		target_is_directory = 1;
+		target_is_directory = true;
 	}
 
 	if (target_is_directory) {
@@ -346,9 +299,10 @@ main(int argc, char* argv[])
 		if (opts_target_directory)
 			last++;
 
-		for (i = first; i < last; i++) {
-			if (stat(argv_copy[i], &s) != 0) {
-				err(1, "failed to stat %s", argv_copy[i]);
+		for (int i = first; i < last; i++) {
+			struct stat s;
+			if (stat(argv[i], &s) != 0) {
+				err(1, "failed to stat %s", argv[i]);
 				return EXIT_FAILURE;
 			}
 			/* We reproduce install's behavior and skip
@@ -358,12 +312,12 @@ main(int argc, char* argv[])
 			if (S_ISDIR(s.st_mode))
 				continue;
 
-			path = path_join(target, basename(argv_copy[i]));
-			copyxattr(argv_copy[i], path);
+			char *path = path_join(target, basename(argv[i]));
+			copyxattr(argv[i], path);
 			free(path);
 		}
 	} else
-		copyxattr(argv_copy[first], target);
+		copyxattr(argv[first], target);
 
 
  done:
